class AutoKeyworder extends BaseConfigPlugin
	getFieldDefFromParm: (baseConfig, fieldName, def) ->
		sort = (a, b) ->
			return a.text.localeCompare(b.text)

		switch def.plugin_type
			when "auto_keyworder_objecttype"
				field = new ez5.ObjecttypeSelector
					form: label: $$("server.config.parameter.system.auto_keyworder.configs.objecttype.label")
					name: fieldName
					show_name: true
					store_value: "fullname"
					filter: (objecttype) =>
						id = objecttype.table.id()
						if CUI.util.isString(id)
							return false
						mask = Mask.getMaskByMaskName("_all_fields", objecttype.table.id())
						if not mask
							return false

						objecttype.addMask(mask)

						data = objecttype: objecttype.table.name()
						if not objecttype.getFields().some((field) => @__filterByAssetField(field, data)) or
							not objecttype.getFields().some((field) => @__filterBySubjectField(field, data)) or
							not objecttype.getFields().some((field) => @__filterByDateTimeField(field, data)) or
							not objecttype.getFields().some((field) => @__filterByKeywordField(field, data))
								return false
						return true
			when "auto_keyworder_asset_version"
				options = [
					text: $$("server.config.parameter.system.auto_keyworder.configs.asset_version.placeholder")
					value: null
					disabled: true
				,
					value: "original"
				]
				for version in ez5.session.config.eas?.image?.versions
					options.push(value: version.version)

				field = new CUI.Select
					form: label: $$("server.config.parameter.system.auto_keyworder.configs.asset_version.label")
					name: fieldName
					options: options
			when "auto_keyworder_asset_field"
				field = new ez5.FieldSelector
					form: label: $$("server.config.parameter.system.auto_keyworder.configs.asset_field.label")
					name: fieldName
					objecttype_data_key: "objecttype"
					store_value: "fullname"
					show_name: true
					placeholder: $$("server.config.parameter.system.auto_keyworder.configs.no_field_selected")
					filter: @__filterByAssetField
					sort: sort
			when "auto_keyworder_subject_field"
				field = new ez5.FieldSelector
					form: label: $$("server.config.parameter.system.auto_keyworder.configs.subject_field.label")
					name: fieldName
					objecttype_data_key: "objecttype"
					store_value: "fullname"
					show_name: true
					placeholder: $$("server.config.parameter.system.auto_keyworder.configs.no_field_selected")
					filter: @__filterBySubjectField
					sort: sort
			when "auto_keyworder_keyword_categories_field"
				field = new ez5.FieldSelector
					form: label: $$("server.config.parameter.system.auto_keyworder.configs.categories_field.label")
					name: fieldName
					objecttype_data_key: "objecttype"
					store_value: "fullname"
					show_name: true
					deep_linked_objects: true
					placeholder: $$("server.config.parameter.system.auto_keyworder.configs.no_field_selected")
					filter: @__filterByKeywordField
					sort: sort
			when "auto_keyworder_keyword_quantity_field"
				field = new ez5.FieldSelector
					form: label: $$("server.config.parameter.system.auto_keyworder.configs.quantity_field.label")
					name: fieldName
					objecttype_data_key: "objecttype"
					store_value: "fullname"
					show_name: true
					deep_linked_objects: true
					placeholder: $$("server.config.parameter.system.auto_keyworder.configs.no_field_selected")
					filter: @__filterByKeywordField
					sort: sort
			when "auto_keyworder_keyword_gender_field"
				field = new ez5.FieldSelector
					form: label: $$("server.config.parameter.system.auto_keyworder.configs.gender_field.label")
					name: fieldName
					objecttype_data_key: "objecttype"
					store_value: "fullname"
					show_name: true
					deep_linked_objects: true
					placeholder: $$("server.config.parameter.system.auto_keyworder.configs.no_field_selected")
					filter: @__filterByKeywordField
					sort: sort
			when "auto_keyworder_keyword_material_field"
				field = new ez5.FieldSelector
					form: label: $$("server.config.parameter.system.auto_keyworder.configs.material_field.label")
					name: fieldName
					objecttype_data_key: "objecttype"
					store_value: "fullname"
					show_name: true
					deep_linked_objects: true
					placeholder: $$("server.config.parameter.system.auto_keyworder.configs.no_field_selected")
					filter: @__filterByKeywordField
					sort: sort
			when "auto_keyworder_keyword_color_field"
				field = new ez5.FieldSelector
					form: label: $$("server.config.parameter.system.auto_keyworder.configs.color_field.label")
					name: fieldName
					objecttype_data_key: "objecttype"
					store_value: "fullname"
					show_name: true
					deep_linked_objects: true
					placeholder: $$("server.config.parameter.system.auto_keyworder.configs.no_field_selected")
					filter: @__filterByKeywordField
					sort: sort
			when "auto_keyworder_timestamp_field"
				field = new ez5.FieldSelector
					form: label: $$("server.config.parameter.system.auto_keyworder.configs.timestamp_field.label")
					name: fieldName
					objecttype_data_key: "objecttype"
					store_value: "fullname"
					show_name: true
					placeholder: $$("server.config.parameter.system.auto_keyworder.configs.no_field_selected")
					filter: @__filterByDateTimeField
					sort: sort
		return field

	__filterByAssetField: (field, data) ->
		return field instanceof EASColumn and not field.insideNested() and field.table.name() == data.objecttype

	__filterBySubjectField: (field, data) ->
		return field instanceof TextColumn and
			field not instanceof NumberColumn and
			not field.insideNested() and field.table.name() == data.objecttype

	__filterByKeywordField: (field, data, parentField) ->
		isKeywordField = field instanceof TextColumn and
			field not instanceof NumberColumn and
			field.table.name() == data.objecttype
		if isKeywordField
			return true

		if parentField instanceof NestedTable and not parentField.insideNested()
			return field instanceof TextColumn and
				field not instanceof NumberColumn

		if parentField instanceof LinkedObject and parentField.insideNested()
			return field instanceof TextColumn and
				field not instanceof NumberColumn and
				not field.insideNested()

		return false

	__filterByDateTimeField: (field, data) ->
		return field instanceof DateTimeColumn and not field.insideNested() and field.table.name() == data.objecttype

ez5.session_ready =>
	BaseConfig.registerPlugin(new AutoKeyworder())